/*
  media-filesystem library, io routines
  tridge@samba.org, January 2001
  released under the Gnu GPL v2
*/

#include "mfs.h"

static int readahead_enabled;
static int vserver = -1;
static int need_bswap;
static int verbose;
static u32 total_size;
static char *dev_list;
extern char *mfs_dev;

void mfs_readahead(int set)
{
	readahead_enabled = set;
}

static void vserver_read_req(u32 sec, u32 count)
{
	struct vserver_cmd cmd;

	cmd.command = htonl(MFS_CMD_READ);
	cmd.param1 = htonl(sec);
	cmd.param2 = htonl(count);
	write_all(vserver, &cmd, sizeof(cmd));
}

static void vserver_receive(void *buf, u32 count)
{
	count <<= SECTOR_SHIFT;
	read_all(vserver, buf, count);
}

#define RA_BLOCKS	256
#define RA_MIN		256

static void vserver_read_sectors(void *buf, u32 sec, u32 count)
{
	static struct mfs_run readahead;
	u32 discard, coming;

	if (count == 0)
		return;

	discard = coming = 0;
	if (sec >= readahead.start && sec < readahead.start + readahead.len) {
		discard = sec - readahead.start;
		coming = readahead.len - discard;
		if (coming <= count) {
			readahead.len = 0;
		} else {
			readahead.start = sec + count;
			readahead.len -= discard + count;
		}
	} else {
		discard = readahead.len;
		coming = 0;
		readahead.len = 0;
	}

	if (coming < count) {
		u32 nreq = count - coming;
		if (readahead_enabled && nreq < RA_BLOCKS) {
			readahead.start = sec + count;
			readahead.len = RA_BLOCKS - nreq;
			nreq = RA_BLOCKS;
		}
		vserver_read_req(sec + coming, nreq);
	}

	if (readahead.len <= RA_MIN && readahead_enabled) {
		if (readahead.len == 0)
			readahead.start = sec + count;
		vserver_read_req(readahead.start + readahead.len, RA_BLOCKS);
		readahead.len += RA_BLOCKS;
	}

	if (discard) {
		void *buf2 = malloc(discard << SECTOR_SHIFT);
		vserver_receive(buf2, discard);
		free(buf2);
	}

	vserver_receive(buf, count);

	if (need_bswap) {
		u16 *v = (u16 *)buf;
		int i;
		for (i=0;i<count<<(SECTOR_SHIFT-1);i++)
			v[i] = ntohs(v[i]);
	}
}

static void vserver_write_sectors(void *buf, u32 sec, u32 count)
{
	struct vserver_cmd cmd;

	if (readahead_enabled) {
		printf("vserver write not supported with readahead\n");
		exit(1);
	}

	cmd.command = htonl(MFS_CMD_WRITE);
	cmd.param1 = htonl(sec);
	cmd.param2 = htonl(count);
	write_all(vserver, &cmd, sizeof(cmd));
	write_all(vserver, buf, count*SECTOR_SIZE);
}

static void vserver_zero_sectors(u32 sec, u32 count)
{
	struct vserver_cmd cmd;

	if (readahead_enabled) {
		printf("vserver write not supported with readahead\n");
		exit(1);
	}

	cmd.command = htonl(MFS_CMD_ZERO);
	cmd.param1 = htonl(sec);
	cmd.param2 = htonl(count);
	write_all(vserver, &cmd, sizeof(cmd));
}

/* this holds the list of block devices in the tivo */
static struct {
	char *dev;
	unsigned long sectors;
	int fd;
} devs[MAX_DEVS];

/* read from the virtual tivo disk (ie. all partitions 
   concatenated) */
void mfs_read_sectors(void *buf, u32 sec, u32 count)
{
	int i;
	loff_t start=0;
	
	if (vserver != -1) {
		vserver_read_sectors(buf, sec, count);
		return;
	}

	sec = partition_remap(sec);

	for (i=0; devs[i].dev; i++) {
		if (sec < start + devs[i].sectors) break;
		start += devs[i].sectors;
	}
	if (!devs[i].dev) {
		fprintf(stderr,"Failed to map sector %d\n", sec);
		exit(1);
	}

	if (verbose) {
		printf("mapped %d to %s/%d\n", sec, devs[i].dev, (int)(sec-start));
	}

	read_sectors(devs[i].fd, buf, sec-start, count);

	if (need_bswap) {
		u16 *v = (u16 *)buf;
		for (i=0;i<count<<(SECTOR_SHIFT-1);i++) v[i] = ntohs(v[i]);
	}
}


/* write to the virtual tivo disk (ie. all partitions 
   concatenated) */
void mfs_write_sectors(void *buf, u32 sec, u32 count)
{
	int i;
	loff_t start=0;
	
	if (vserver != -1) {
		vserver_write_sectors(buf, sec, count);
		return;
	}

	sec = partition_remap(sec);

	for (i=0; devs[i].dev; i++) {
		if (sec < start + devs[i].sectors) break;
		start += devs[i].sectors;
	}
	if (!devs[i].dev) {
		fprintf(stderr,"Failed to map sector %d\n", sec);
		exit(1);
	}

	if (verbose) {
		printf("mapped %d to %s/%d\n", sec, devs[i].dev, (int)(sec-start));
	}

	write_sectors(devs[i].fd, buf, sec-start, count);
}


void mfs_zero_sectors(int sector, int count)
{
	int chunk_size=512;
	char buf[chunk_size*512];
	char buf1[chunk_size*512];
	int total=0;

	if (vserver != -1) {
		vserver_zero_sectors(sector, count);
		return;
	}

	bzero(buf, chunk_size*512);

	printf("wiping %d sectors at %d\n", count, sector);

	while (total<count) {
		int n = MIN(chunk_size, count);
		mfs_read_sectors(buf1, sector, n);
		if (memcmp(buf, buf1, n*SECTOR_SIZE)) {
			mfs_write_sectors(buf, sector, n);
		}
		sector += n;
		total += n;
		printf("%3.1f%%\r", 100.0*total/count);
		fflush(stdout);
	}
	printf("%3.1f%%\n", 100.0*total/count);
}


/* initialise the devices list from the superblock devlist */
void load_devs(char *devlist)
{
	char *p;
	int i=0, len;

	if (dev_list) devlist = dev_list;

	if (vserver != -1) return;

	total_size = 0;

	while (1) {
		p = strchr(devlist, ' ');
		if (!p) {
			len = strlen(devlist);
		} else {
			len = (int)(p-devlist);
		}
		devs[i].dev = (char *)malloc(len+1);
		strncpy(devs[i].dev, devlist, len);
		devs[i].dev[len] = 0;
		devs[i].fd = open(devs[i].dev, O_RDWR|O_LARGEFILE);
		if (devs[i].fd == -1) {
			printf("failed to open [%s]\n", devs[i].dev);
			break;
		}
		devs[i].sectors = llseek(devs[i].fd, 0, SEEK_END) >> SECTOR_SHIFT;
		if (devs[i].sectors == 0) {
			ioctl(devs[i].fd, BLKGETSIZE, &devs[i].sectors);
		}
		devs[i].sectors &= ~(MFS_BLOCK_ROUND-1);
		total_size += devs[i].sectors;
		i++;
		devlist += len+1;
		if (!p) break;
	}	
}

void add_dev_map(char *mapping)
{
	char *p;

	if (mapping[0] == ':') {
		vserver = open_socket_out(mapping+1, VSERVER_PORT);
		if (vserver == -1) {
			fprintf(stderr,"Failed to connect to %s\n", mapping+1);
			exit(1);
		}
		return;
	}

	dev_list = strdup(mapping);
	mfs_dev = strdup(dev_list);
	if ((p = strchr(mfs_dev, ' '))) *p = 0;
}

void io_dev_info(void)
{
	int i;

	for (i=0; devs[i].dev; i++) {
		printf("%s has %ld sectors\n", 
		       devs[i].dev,
		       devs[i].sectors);
	}
}

int io_vserver(void)
{
	return vserver;
}

void io_need_bswap(int set)
{
	need_bswap = set;
}

u32 io_total_size(void)
{
	if (partition_total_size()) {
		total_size = partition_total_size();
	}	
	return total_size;
}


/* read bytes from a sector, handling partial sectors */
void mfs_read_partial(void *buf, u32 sec, u32 size)
{
	char tmp[SECTOR_SIZE];
	if (size >= SECTOR_SIZE) {
		u32 n = size / SECTOR_SIZE;
		mfs_read_sectors(buf, sec, n);
		buf += n*SECTOR_SIZE;
		size -= n*SECTOR_SIZE;
		sec += n;
	}
	if (size == 0) return;
	mfs_read_sectors(tmp, sec, 1);
	memcpy(buf, tmp, size);
}
