/*
  media-filesystem object schema code
  tridge@samba.org, January 2001
  released under the Gnu GPL v2
*/

#include "mfs.h"

#define MAX_TYPES 200
#define MAX_ATTRS 70

static int current_type;
static char *types[MAX_TYPES];

static struct {
	char *name;
	int objtype;
} attrs[MAX_TYPES][MAX_ATTRS];

static void schema_add(int type, int attr, char *name, int objtype)
{
	int i;
	name = strdup(name);
	if (type != -1) {
		attrs[type][attr].name = name;
		attrs[type][attr].objtype = objtype;
		return;
	}
	for (i=0;i<MAX_TYPES;i++) {
		attrs[i][attr].name = name;
		attrs[i][attr].objtype = objtype;
	}
	// printf("added %d/%s\n", type, name);
}

/* used to load a local schema.txt to make things faster */
static int preload_schema(char *fname)
{
	FILE *f = fopen(fname, "r");
	int itype, iattr, atype=0;
	char *type, *attr, *flag;

	char line[200];

	if (!f) return 0;

	while (fgets(line, sizeof(line), f)) {
		if (!isdigit(line[0])) continue;
		if (line[strlen(line)-1] == '\n') line[strlen(line)-1] = 0;
		itype = atoi(strtok(line,"\t "));
		type = strtok(NULL,"\t ");
		iattr = atoi(strtok(NULL,"\t "));
		attr = strtok(NULL,"\t ");
		flag = strtok(NULL,"\t ");

		if (!types[itype]) types[itype] = strdup(type);
		attrs[itype][iattr].name = strdup(attr);
		if (strcmp(flag,"string")==0) {
			atype = TYPE_STRING;
		} else if (strcmp(flag,"int")==0) {
			atype = TYPE_INT;
		} else if (strcmp(flag,"object")==0) {
			atype = TYPE_OBJECT;
		} else if (strcmp(flag,"file")==0) {
			atype = TYPE_FILE;
		}
		if (!types[itype]) types[itype] = strdup(type);
		schema_add(itype, iattr, attr, atype);
		// printf("preloaded %d/%d/%s/%d\n",itype, iattr, attr, atype);
	}

	fclose(f);
	return 1;
}

static void load_types(void)
{
	struct mfs_dirent *dir;
	u32 count, i;
	static int loaded;
	
	if (loaded) return;

	loaded = 1;

	if (preload_schema("schema.txt")) return;

	types[40] = "ObjectType";
	
	dir = mfs_dir(mfs_resolve("/ObjectType"), &count);

	for (i=0;i<count;i++) {
		int index = query_int(dir[i].fsid, "Index");
		if (!types[index]) {
			// printf("adding type %d %s\n", index, dir[i].name);
			types[index] = strdup(dir[i].name);
		}
	}

	if (dir) mfs_dir_free(dir);
}


static void load_schema(int type)
{
	u32 size, fsid;
	char path[100];
	void *buf;

	void load_callback2(int fsid, struct mfs_subobj_header *obj,
			    struct mfs_attr_header *attr, void *data)
		{
			static char *name;
			static u32 iattr;
			if (obj->flags || !attr) return;
			if (attr->attr == 16) {
				name = (char *)data;
			} else if (attr->attr == 17) {
				iattr = ntohl(*(u32 *)data);
			} else if (attr->attr == 18) {
				u32 objtype = ntohl(*(u32 *)data);
				schema_add(current_type, iattr, name, objtype);
			}
		}

	void load_callback1(int fsid, struct mfs_subobj_header *obj,
			    struct mfs_attr_header *attr, void *data)
		{
			static char *name;
			if (!obj->flags || !attr) return;
			if (attr->attr == 16) {
				name = strdup((char *)data);
			} else if (attr->attr == 17) {
				current_type = ntohl(*(u32 *)data);
				types[current_type] = name;
			}
		}


	if (!attrs[0][1].name) {
		schema_add(-1, 1, "Version", TYPE_INT);
		schema_add(-1, 2, "Expiration", TYPE_INT);
		schema_add(-1, 3, "Path", TYPE_STRING);
		schema_add(-1, 4, "IndexPath", TYPE_STRING);
		schema_add(-1, 5, "IndexUsed", TYPE_OBJECT);
		schema_add(-1, 6, "IndexUsedBy", TYPE_OBJECT);
		schema_add(-1, 7, "IndexAttr", TYPE_OBJECT);
		schema_add(-1, 8, "ServerId", TYPE_STRING);
		schema_add(-1, 9, "ServerVersion", TYPE_INT);
		schema_add(-1, 10, "Uuid", TYPE_STRING);
		schema_add(-1, 11, "Unsatisfied", TYPE_STRING);
		schema_add(-1, 12, "Bits", TYPE_INT);
		schema_add(-1, 13, "Unused1", TYPE_INT);
		schema_add(-1, 14, "Unused2", TYPE_INT);
		schema_add(-1, 15, "Unused3", TYPE_INT);
	}

	sprintf(path, "/ObjectType/%s", schema_type(type));

	fsid = mfs_resolve(path);
	size = mfs_fsid_size(fsid);
	if (size > 0) {
		buf = malloc(size);
		mfs_fsid_pread(fsid, buf, 0, size);
		parse_object(fsid, buf, load_callback1);
		parse_object(fsid, buf, load_callback2);
		free(buf);
	}
}

/* lookup a string for a schema type */
char *schema_type(int type)
{
	if (!types[type]) load_types();

	return types[type];
}

/* lookup an attribute for a given type and attr value,
   auto-loading the schema if necessary */
char *schema_attrib(int type, int attr)
{
	if (!attrs[type][attr].name) {
		load_schema(type);
	}
	if (!attrs[type][attr].name) {
		char name[30];
		snprintf(name, sizeof(name)-1, "UNKNOWN(%d)", attr);
		attrs[type][attr].name = strdup(name);
	}
	//	printf("schema_attrib(%d, %d) -> %s\n", type, attr, attrs[type][attr]);
	return attrs[type][attr].name;
}

/* dump the complete schema */
void dump_schema(FILE *f)
{
	int i, j;

	load_types();

	for (i=1; i<MAX_TYPES; i++) {
		if (!types[i]) continue;
		load_schema(i);
		for (j=1; j<MAX_ATTRS; j++) {
			if (attrs[i][j].name) {
				printf("%d %s %d %s %s\n",
				       i, types[i], j, attrs[i][j].name,
				       object_typestr(attrs[i][j].objtype));
			}
		}
	}
}

